import numpy as np

from scipy.signal import butter, sosfiltfilt, ricker, cwt, correlate, savgol_filter
from scipy.optimize import curve_fit, differential_evolution

import matplotlib.pyplot as plt
from matplotlib.colors import colorConverter
import matplotlib.transforms as transforms
import matplotlib.patches as patches

EXCLUDE_COLOR = 'xkcd:salmon'
SAMPLE_SETTINGS = {
    "Glucose [mM]": 8,
    "Sampling [Hz]": 10,
    "Stimulation [frame]": [0, 0],
    "Filter":
        {
        "Slow [Hz]": [0.001, 0.005],
        "Fast [Hz]": [0.04, 0.4],
        "Plot [s]": [250, 1750]
        },
    "Distribution order": 5,
    "Exclude":
        {
        "Score threshold": 1.5,
        "Spikes threshold": 0.01
        },
    "Distance [um]": 1
    }

class Data(object):
    """
    A class for signal analysis.
    """
# ------------------------------- INITIALIZER -------------------------------- #
    def __init__(self):
        self.__signal = False
        self.__mean_islet = False
        self.__time = False
        self.__settings = False

        self.__points = False
        self.__cells = False
        self.__filtered_slow = False
        self.__filtered_fast = False
        self.__distributions = False
        self.__binarized_slow = False
        self.__binarized_fast = False
        self.__activity = False
        self.__good_cells = False

# --------------------------------- IMPORTS ---------------------------------- #
    def import_data(self, signal):
        if not len(signal.shape)==2:
            raise ValueError("Signal shape not 2D.")
        self.__signal = np.around(signal[:,1:].transpose(), decimals=3)
        self.__mean_islet = np.mean(self.__signal, 0) # average over 0 axis
        self.__mean_islet = self.__mean_islet - np.mean(self.__mean_islet)
        if self.__settings is False:
            self.__settings = SAMPLE_SETTINGS
        sampling = self.__settings["Sampling [Hz]"]
        self.__time = np.arange(len(self.__signal[0]))*(1/sampling)

        self.__points = len(self.__time)
        self.__cells = len(self.__signal)

        self.__good_cells = np.ones(self.__cells, dtype="bool")

    def import_settings(self, settings=SAMPLE_SETTINGS):
        if not "Sampling [Hz]" in settings and not "Stimulation [frame]" in settings and not "Filter" in settings and not "Exclude" in settings and not "Distance [um]" in settings:
            raise ValueError("Bad keys in settings.")
        if not "Slow [Hz]" in settings["Filter"] and not "Fast [Hz]" in settings["Filter"] and not "Plot [s]" in settings["Filter"]:
            raise ValueError("Bad keys in settings[filter].")
        if not "Score threshold" in settings["Exclude"] and not "Spikes threshold" in settings["Exclude"]:
            raise ValueError("Bad keys in settings[exclude].")
        self.__settings = settings

    def import_good_cells(self, cells):
        if self.__signal is False:
            raise ValueError("No imported data!")
        if len(cells) != self.__cells:
            raise ValueError("Cell number does not match.")
        self.__good_cells = cells

    def reset_computations(self):
        self.__filtered_slow = False
        self.__filtered_fast = False
        self.__distributions = False
        self.__binarized_slow = False
        self.__binarized_fast = False
        self.__activity = False
        self.__good_cells = np.ones(self.__cells, dtype="bool")


# --------------------------------- GETTERS ---------------------------------- #
    def get_settings(self): return self.__settings
    def get_time(self): return self.__time
    def get_signal(self): return self.__signal
    def get_mean_islet(self): return self.__mean_islet
    def get_points(self): return self.__points
    def get_cells(self): return self.__cells
    def get_filtered_slow(self): return self.__filtered_slow
    def get_filtered_fast(self): return self.__filtered_fast
    def get_distributions(self): return self.__distributions
    def get_binarized_slow(self): return self.__binarized_slow
    def get_binarized_fast(self): return self.__binarized_fast
    def get_activity(self): return self.__activity
    def get_good_cells(self): return self.__good_cells

    def plot(self, ax, cell, plots=("mean", "raw", "slow", "fast"), protocol=True):
        time = self.__time
        sampling = self.__settings["Sampling [Hz]"]
        glucose = self.__settings["Glucose [mM]"]
        TA, TAE = self.__settings["Stimulation [frame]"]
        TA, TAE = TA/sampling, TAE/sampling

        if "mean" in plots:
            signal = self.__mean_islet
            signal = signal/np.max(signal)
            ax.plot(time, signal, "k", alpha=0.25, lw=0.1)
        if "raw" in plots:
            signal = self.__signal[cell]
            signal = signal - np.mean(signal)
            signal = signal/np.max(signal)
            ax.plot(time, signal, "k", alpha=0.5, lw=0.1)
        if "slow" in plots:
            filtered_slow = self.__filtered_slow[cell]
            filtered_slow = filtered_slow/np.max(filtered_slow)
            ax.plot(time, filtered_slow, color="C0", lw=2)
        if "fast" in plots:
            filtered_fast = self.__filtered_fast[cell]
            filtered_fast = filtered_fast/np.max(filtered_fast)
            ax.plot(time, filtered_fast, color="C3", lw=0.2)
        if "bin_slow" in plots:
            binarized_slow = self.__binarized_slow[cell]
            ax2 = ax.twinx()
            ax2.plot(time, binarized_slow, color="k", lw=1)
            ax2.set_ylabel("Phase")
        if "bin_fast" in plots:
            threshold = self.__distributions[cell]["p_root"]
            filtered_fast = self.__filtered_fast[cell]
            binarized_slow = self.__binarized_fast[cell]/np.max(filtered_fast)*threshold
            ax.plot(time, binarized_slow, color="k", lw=1)
            ax2 = ax.twinx()
            ax2.set_ylabel("Action potentials")

        ax.set_xlim(0,self.__time[-1])
        ax.set_ylim(None, 1.1)
        ax.set_xlabel("Time [s]")
        ax.set_ylabel("Amplitude")

        ax.axvline(self.__settings["Stimulation [frame]"][0]/self.__settings["Sampling [Hz]"], c="grey")
        ax.axvline(self.__settings["Stimulation [frame]"][1]/self.__settings["Sampling [Hz]"], c="grey")

        if self.__good_cells[cell]:
            if self.__activity is not False:
                border = self.__activity[cell]
                ax.axvspan(0, border[0], alpha=0.25, color="grey")
                ax.axvspan(border[1], self.__time[-1], alpha=0.25, color="grey")
        else:
            ax.axvspan(0,self.__time[-1], alpha=0.5, color=EXCLUDE_COLOR)

        if protocol and TA!=0 and TAE!=0:
            color = "C0" if glucose == 8 else "C3"
            # tform = transforms.blended_transform_factory(ax.transData, ax.transAxes)
            rectangles = {'' : patches.Rectangle((0, 1.1), TA, 0.15, color ='grey', alpha=0.5, transform=ax.transData, clip_on=False),
                          '{} mM'.format(glucose) : patches.Rectangle((TA, 1.1), TAE-TA, 0.3, color=color, alpha=0.8, transform=ax.transData, clip_on=False),
                          '6 mM' : patches.Rectangle((TAE, 1.1), time[-1]-TAE, 0.15, color ='grey', alpha=0.5, transform=ax.transData, clip_on=False)
                         }
            for r in rectangles:
                ax.add_artist(rectangles[r])
                rx, ry = rectangles[r].get_xy()
                cx = rx + rectangles[r].get_width()/2.0
                cy = ry + rectangles[r].get_height()/2.0
                ax.annotate(r, (cx, cy), color='k', fontsize=12, ha='center', va='center', xycoords=ax.transData, annotation_clip=False)

# ----------------------------- ANALYSIS METHODS ----------------------------- #
    def plot_raw(self, i):
        if self.__signal is False:
            raise ValueError("No imported data!")
        if i not in range(self.__cells):
            raise ValueError("Cell index not in range.")
        mean = np.mean(self.__signal[i])

        fig, (ax1, ax2) = plt.subplots(2, sharex=True)
        self.plot(ax1, i, plots=("mean"))
        self.plot(ax2, i, plots=("raw"), protocol=False)

        return fig
# ---------- Filter + smooth ---------- #
    def filter(self):
        if self.__signal is False:
            raise ValueError("No imported data!")
        slow, fast = self.__settings["Filter"]["Slow [Hz]"], self.__settings["Filter"]["Fast [Hz]"]
        self.__filtered_slow = np.zeros((self.__cells, self.__points))
        self.__filtered_fast = np.zeros((self.__cells, self.__points))

        for i in range(self.__cells):
            self.__filtered_slow[i] = self.__bandpass(self.__signal[i], (*slow))
            self.__filtered_fast[i] = self.__bandpass(self.__signal[i], (*fast))

    def __bandpass(self, data, lowcut, highcut, order=5):
        nyq = 0.5*self.__settings["Sampling [Hz]"]
        low = lowcut / nyq
        high = highcut / nyq
        sos = butter(order, [low, high], analog=False, btype='band', output='sos')
        y = sosfiltfilt(sos, data)
        return y

    def plot_filtered(self, i):
        if self.__filtered_slow is False or self.__filtered_fast is False:
            raise ValueError("No filtered data!")
        if i not in range(self.__cells):
            raise ValueError("Cell index not in range.")

        fig, (ax1, ax2) = plt.subplots(2)
        fig.suptitle("Filtered data")

        self.plot(ax1, i, plots=("raw, slow"))
        ax1.set_xlabel(None)

        self.plot(ax2, i, plots=("raw, fast"), protocol=False)
        ax2.set_xlim(*self.__settings["Filter"]["Plot [s]"])

        return fig

# ---------- Binarize ---------- #

    def compute_distributions(self):
        if self.__filtered_slow is False:
            raise ValueError("No filtered data.")
        self.__distributions = [dict() for i in range(self.__cells)]

        for cell in range(self.__cells):
            # Compute cumulative histogram and bins
            signal = np.clip(self.__filtered_fast[cell], 0, None)
            hist = np.histogram(signal, 50)
            cumulative_hist = np.flip(np.cumsum(np.flip(hist[0])))
            bins = hist[1]
            x = (bins[1:] + bins[:-1])/2 # middle points of bins

            # Fit polynomial of nth order
            order = self.__settings["Distribution order"]
            z = np.polyfit(x, np.log(cumulative_hist), order, w=np.sqrt(cumulative_hist))
            p = np.poly1d(z)

            # Transitions of polynomial
            first_derivative = np.polyder(p, 1)
            second_derivative = np.polyder(p, 2)
            roots = np.roots(second_derivative)
            real_roots = roots.real[abs(roots.imag) < 1e-5]
            final_roots = real_roots[np.logical_and(real_roots > 0, real_roots < bins[-1])]
            p_root = min(final_roots)
            min_value = first_derivative(p_root)

            # Quadratic function around 0
            q = np.poly1d([second_derivative(0)/2, first_derivative(0), p(0)])
            q_root = np.roots(np.polyder(q, 1))
            q_root = q_root[0] if q_root>0 or q_root>bins[-1] else np.inf

            # Goodness score
            phi = np.arctan(abs(min_value))/(np.pi/2)
            score = (1-phi)*p_root/q_root

            spikes_threshold = self.__settings["Exclude"]["Spikes threshold"]
            if np.exp(p(p_root))<spikes_threshold*self.__points:
                score = 0

            self.__distributions[cell]["hist"] = cumulative_hist
            self.__distributions[cell]["bins"] = x
            self.__distributions[cell]["p"] = p
            self.__distributions[cell]["q"] = q
            self.__distributions[cell]["p_root"] = p_root
            self.__distributions[cell]["q_root"] = q_root
            self.__distributions[cell]["score"] = score

    def plot_distributions(self, i):
        if self.__distributions is False:
            raise ValueError("No distribution data.")
        hist = self.__distributions[i]["hist"]
        x = self.__distributions[i]["bins"]
        p = self.__distributions[i]["p"]
        q = self.__distributions[i]["q"]
        p_root = self.__distributions[i]["p_root"]
        q_root = self.__distributions[i]["q_root"]
        score = self.__distributions[i]["score"]

        fig, (ax1, ax2) = plt.subplots(2, 1)
        fig.suptitle("Distribution (score = {0:.2f})".format(score), x=0.01, y=.98, ha="left")

        ax1.bar(x, hist, max(x)/len(x)*0.8, log=True, color="dimgrey", alpha=0.5)
        ax1.plot(x, np.exp(p(x)), color="red")
        ax1.plot(p_root, np.exp(p(p_root)), marker="o", color="k")
        if q_root != np.inf:
            ax1.plot(np.linspace(0,q_root,10), np.exp(q(np.linspace(0,q_root,10))), color=EXCLUDE_COLOR)
            ax1.plot(q_root, np.exp(q(q_root)), marker="o", color=EXCLUDE_COLOR)
        ax1.set_xlabel("Signal height h")
        ax1.set_ylabel("Data points N")

        ax1.xaxis.tick_top()
        ax1.xaxis.set_label_position('top')

        self.plot(ax2, i, plots=("raw, fast"), protocol=False)
        scaling = np.max(self.__filtered_fast[i])
        ax2.axhline(p_root/scaling, c="k")
        if q_root is not np.inf:
            ax2.fill_between(self.__time, -q_root/scaling, q_root/scaling, color=EXCLUDE_COLOR, zorder=1, alpha=0.5)

        return fig

# ---------- Exclude ---------- #
    def autoexclude(self):
        if self.__distributions is False:
            raise ValueError("No distributions.")
        # Excluding thresholds
        score_threshold = self.__settings["Exclude"]["Score threshold"]

        for cell in range(self.__cells):
            p = self.__distributions[cell]["p"]
            p_root = self.__distributions[cell]["p_root"]
            score = self.__distributions[cell]["score"]
            if score<score_threshold:
                self.__good_cells[cell] = False

    def exclude(self, cell):
        if cell not in range(self.__cells):
            raise ValueError("Cell not in range.")
        self.__good_cells[cell] = False

    def unexclude(self, cell):
        if cell not in range(self.__cells):
            raise ValueError("Cell not in range.")
        self.__good_cells[cell] = True

# ---------- Binarize ---------- #

    def __search_sequence(self, arr, seq):
        # Store sizes of input array and sequence
        seq = np.array(seq)
        Na, Nseq = arr.size, seq.size

        # Range of sequence
        r_seq = np.arange(Nseq)

        # Create a 2D array of sliding indices across the entire length of input array.
        # Match up with the input sequence & get the matching starting indices.
        M = (arr[np.arange(Na-Nseq+1)[:,None] + r_seq] == seq).all(1)

        # Get the range of those indices as final output
        if M.any() > 0:
            return np.where(M == True)[0]
        else:
            return np.array([], dtype="int") # No match found

    def binarize_fast(self):
        if self.__distributions is False or self.__filtered_fast is False:
            raise ValueError("No distribution or filtered data.")
        self.__binarized_fast = np.zeros((self.__cells, self.__points))
        for cell in range(self.__cells):
            threshold = self.__distributions[cell]["p_root"]
            self.__binarized_fast[cell] = np.where(self.__filtered_fast[cell]>threshold, 1, 0)
            self.__binarized_fast = self.__binarized_fast.astype(int)

    def binarize_slow(self):
        if self.__filtered_slow is False:
            raise ValueError("No filtered data.")
        self.__binarized_slow = np.zeros((self.__cells, self.__points))
        for cell in range(self.__cells):
            signal = self.__filtered_slow[cell]
            heavisided_gradient = np.heaviside(np.gradient(signal), 0)
            minima = self.__search_sequence(heavisided_gradient, [0,1])
            maxima = self.__search_sequence(heavisided_gradient, [1,0])
            extremes = np.sort(np.concatenate((minima, maxima)))

            reverse_mode = False if minima[0] < maxima[0] else True

            self.__binarized_slow[cell, 0:extremes[0]] = 0
            for i in range(len(extremes)-1):
                e1, e2 = extremes[i], extremes[i+1]
                if i%2 == int(reverse_mode):
                    self.__binarized_slow[cell, e1:e2] = np.floor(np.linspace(1,7,e2-e1, endpoint=False))
                else:
                    self.__binarized_slow[cell, e1:e2] = np.floor(np.linspace(7,13,e2-e1, endpoint=False))
            self.__binarized_slow[cell, extremes[-1]:] = 0
        self.__binarized_slow = self.__binarized_slow.astype(int)

    def autolimit(self):
        if self.__binarized_fast is False:
            raise ValueError("No binarized data.")
        print("Computing activity...")
        self.__activity = []
        for cell in range(self.__cells):
            data = self.__binarized_fast[cell]
            cumsum = np.cumsum(data)

            sampling = self.__settings["Sampling [Hz]"]
            stimulation = self.__settings["Stimulation [frame]"][0]
            lower_limit = cumsum[cumsum<0.1*cumsum[-1]].size/sampling
            upper_limit = (cumsum.size - cumsum[cumsum>0.9*cumsum[-1]].size)/sampling

            box = lambda t, a, t_start, t_end: a*(np.heaviside(t-t_start, 0)-np.heaviside(t-t_end, 0))
            res = differential_evolution(lambda p: np.sum((box(self.__time, *p) - data)**2),  # quadratic cost function
                                 # [[0, 100], [0, t_half], [t_half, 2*t_half]])  # parameter bounds
                                 [[0, 100], [0, lower_limit], [upper_limit, self.__time[-1]]])  # parameter bounds
            self.__activity.append(res.x[1:])

            if self.__activity[cell][0] < stimulation/sampling:
                self.__good_cells[cell] = False
        self.__activity = np.array(self.__activity)

    def plot_binarized(self, cell):
        if self.__binarized_slow is False or self.__binarized_fast is False:
            raise ValueError("No binarized data!")

        fig, (ax1, ax2) = plt.subplots(2, 1, sharex=True)
        fig.suptitle("Binarized data")

        self.plot(ax1, cell, plots=("slow", "bin_slow"))
        self.plot(ax2, cell, plots=("fast", "bin_fast"), protocol=False)

        return fig

    def plot_events(self):
        if self.__binarized_slow is False or self.__binarized_fast is False:
            raise ValueError("No binarized data!")
        fig, ax = plt.subplots()

        bin_fast = self.__binarized_fast[self.__good_cells]
        raster = [[] for i in range(len(bin_fast))]
        sampling = self.__settings["Sampling [Hz]"]

        for i in range(len(bin_fast)):
            for j in range(len(bin_fast[0])):
                if bin_fast[i,j] == 1:
                    raster[i].append(j/sampling)

        ax.eventplot(raster, linewidths=0.1)
        return fig


    def is_analyzed(self):
        if self.__filtered_slow is False or self.__filtered_fast is False:
            return False
        elif self.__distributions is False:
            return False
        elif self.__binarized_slow is False or self.__binarized_fast is False:
            return False
        elif self.__good_cells is False:
            return False
        elif self.__activity is False:
            return False
        else:
            return True



import numpy as np
import matplotlib.pyplot as plt

from .networks import Networks

class Analysis(object):
    """docstring for Analysis."""

# ------------------------------- INITIALIZER -------------------------------- #
    def __init__(self):
        self.__settings = None
        self.__points = None
        self.__cells = None
        self.__positions = None
        self.__filtered_slow = None
        self.__filtered_fast = None
        self.__binarized_slow = None
        self.__binarized_fast = None
        self.__activity = None

        self.__networks = False

    def import_data(self, data, positions):
        assert data.is_analyzed()

        good_cells = data.get_good_cells()

        self.__settings = data.get_settings()
        self.__sampling = self.__settings["Sampling [Hz]"]
        self.__points = data.get_points()
        self.__positions = positions[good_cells]*self.__settings["Distance [um]"]
        self.__cells = np.sum(good_cells)

        self.__filtered_slow = data.get_filtered_slow()[good_cells]
        self.__filtered_fast = data.get_filtered_fast()[good_cells]

        self.__binarized_slow = data.get_binarized_slow()[good_cells]
        self.__binarized_fast = data.get_binarized_fast()[good_cells]

        self.__activity = np.array(data.get_activity())[good_cells]


    def build_networks(self):
        print("Building networks...")
        # Construct networks and build networks from data
        self.__networks = Networks(self.__cells, self.__filtered_slow, self.__filtered_fast)
        self.__networks.build_networks()

# ---------------------------- ANALYSIS FUNCTIONS ---------------------------- #
    def __search_sequence(self, arr, seq):
        # Store sizes of input array and sequence
        seq = np.array(seq)
        Na, Nseq = arr.size, seq.size

        # Range of sequence
        r_seq = np.arange(Nseq)

        # Create a 2D array of sliding indices across the entire length of input array.
        # Match up with the input sequence & get the matching starting indices.
        M = (arr[np.arange(Na-Nseq+1)[:,None] + r_seq] == seq).all(1)

        # Get the range of those indices as final output
        if M.any() > 0:
            return np.where(M == True)[0]
        else:
            return np.array([], dtype="int") # No match found

    def __distances_matrix(self):
        A_dst = np.zeros((self.__cells, self.__cells))
        for cell1 in range(self.__cells):
            for cell2 in range(cell1):
                x1, y1 = self.__positions[cell1,0], self.__positions[cell1,1]
                x2, y2 = self.__positions[cell2,0], self.__positions[cell2,1]
                distance = np.sqrt((x1-x2)**2 + (y1-y2)**2)
                A_dst[cell1, cell2] = distance
                A_dst[cell2, cell1] = distance
        return A_dst

# ------------------------------ GETTER METHODS ------------------------------ #

    def get_positions(self): return self.__positions
    def get_filtered_slow(self): return self.__filtered_slow
    def get_filtered_fast(self): return self.__filtered_fast
    def get_networks(self): return self.__networks

# ----------------------------- PARAMETER METHODS ---------------------------- #

    def draw_networks(self, ax1, ax2, colors):
        return self.__networks.draw_networks(self.__positions, ax1, ax2, colors)

    def compute_parameters(self):

        par_cell = [dict() for c in range(self.__cells)]
        par_network = dict()

        if self.__networks is not False:
            par_network["Rs"] = self.average_correlation()[0]
            par_network["Rf"] = self.average_correlation()[1]
            par_network["Ds"] = self.connection_distances()[0]
            par_network["Df"] = self.connection_distances()[1]
            par_network["Qs"] = self.modularity()[0]
            par_network["Qf"] = self.modularity()[1]
            par_network["GEs"] = self.global_efficiency()[0]
            par_network["GEf"] = self.global_efficiency()[1]
            par_network["MCCs"] = self.max_connected_component()[0]
            par_network["MCCf"] = self.max_connected_component()[1]
        par_network["MA"] = self.mean_amplitude()
        for cell in range(self.__cells):
            par_cell[cell]["AD"] = self.activity(cell)[0]
            par_cell[cell]["AT"] = self.activity(cell)[1]
            par_cell[cell]["OD"] = self.activity(cell)[2]
            par_cell[cell]["Fs"] = self.frequency(cell)[0]
            par_cell[cell]["Ff"] = self.frequency(cell)[1]
            par_cell[cell]["ISI"] = self.interspike(cell)[0]
            par_cell[cell]["ISIV"] = self.interspike(cell)[1]
            par_cell[cell]["TP"] = self.time(cell)["plateau_start"]
            par_cell[cell]["TS"] = self.time(cell)["spike_start"]
            par_cell[cell]["TI"] = self.time(cell)["plateau_end"]

            if self.__networks is not False:
                par_cell[cell]["NDs"] = self.node_degree(cell)[0]
                par_cell[cell]["NDf"] = self.node_degree(cell)[1]
                par_cell[cell]["Cs"] = self.clustering(cell)[0]
                par_cell[cell]["Cf"] = self.clustering(cell)[1]
                par_cell[cell]["NNDs"] = self.nearest_neighbour_degree(cell)[0]
                par_cell[cell]["NNDf"] = self.nearest_neighbour_degree(cell)[1]

        return (par_cell, par_network)

# ----------------------- INDIVIDUAL PARAMETER METHODS ----------------------- #

    def average_correlation(self):
        if self.__networks is False:
            raise ValueError("Network is not built.")
        return self.__networks.average_correlation()

    def connection_distances(self):
        if self.__networks is False:
            raise ValueError("Network is not built.")
        A_dst = self.__distances_matrix()
        A_slow, A_fast = self.__networks.get_A_slow(), self.__networks.get_A_fast()

        A_dst_slow = np.multiply(A_dst, A_slow)
        A_dst_fast = np.multiply(A_dst, A_fast)

        slow_distances, fast_distances = [], []
        for c1 in range(self.__cells):
            for c2 in range(c1):
                ds = A_dst_slow[c1,c2]
                df = A_dst_fast[c1,c2]
                if ds > 0: slow_distances.append(ds)
                if df > 0: fast_distances.append(df)

        return (np.array(slow_distances), np.array(fast_distances))

    def modularity(self):
        if self.__networks is False:
            raise ValueError("Network is not built.")
        return self.__networks.modularity()

    def global_efficiency(self):
        if self.__networks is False:
            raise ValueError("Network is not built.")
        return self.__networks.global_efficiency()

    def max_connected_component(self):
        if self.__networks is False:
            raise ValueError("Network is not built.")
        return self.__networks.max_connected_component()

    def mean_amplitude(self):
        amplitudes = []
        for cell in range(self.__cells):
            heavisided_gradient = np.heaviside(np.gradient(self.__filtered_slow[cell]), 0)
            minima = self.__search_sequence(heavisided_gradient, [0,1])
            maxima = self.__search_sequence(heavisided_gradient, [1,0])

            if maxima[0] < minima[0]: maxima = np.delete(maxima, 0)
            if maxima[-1] < minima[-1]: minima = np.delete(minima, 0)

            for i, j in zip(minima, maxima):
                amplitudes.append(self.__filtered_slow[cell][j]-self.__filtered_slow[cell][i])
        return np.mean(amplitudes)

    def activity(self, cell):
        start = int(self.__activity[cell][0]*self.__sampling)
        stop = int(self.__activity[cell][1]*self.__sampling)
        bin = self.__binarized_fast[cell][start:stop]
        sum = np.sum(bin)
        length = bin.size
        Nf = self.frequency(cell)[1]*length/self.__sampling
        if sum == 0:
            return (length, np.nan, np.nan)
        return (length/self.__sampling, sum/length, (sum/self.__sampling)/Nf)

    def frequency(self, cell):
        start = int(self.__activity[cell][0]*self.__sampling)
        stop = int(self.__activity[cell][1]*self.__sampling)
        bin_slow = self.__binarized_slow[cell][start:stop]
        bin_fast = self.__binarized_fast[cell][start:stop]

        slow_peaks = self.__search_sequence(bin_slow, [11,12])
        if slow_peaks.size < 2:
            frequency_slow = np.nan
        else:
            slow_interval = slow_peaks[-1]-slow_peaks[0]
            frequency_slow = (slow_peaks.size-1)/slow_interval*self.__sampling

        fast_peaks = self.__search_sequence(bin_fast, [0,1])
        if fast_peaks.size < 2:
            frequency_fast = np.nan
        else:
            fast_interval = fast_peaks[-1]-fast_peaks[0]
            frequency_fast = (fast_peaks.size-1)/fast_interval*self.__sampling

        return (frequency_slow, frequency_fast)

    def interspike(self, cell):
        start = int(self.__activity[cell][0]*self.__sampling)
        stop = int(self.__activity[cell][1]*self.__sampling)
        bin_fast = self.__binarized_fast[cell][start:stop]

        interspike_start = self.__search_sequence(bin_fast, [1,0])
        interspike_end = self.__search_sequence(bin_fast, [0,1])

        if interspike_start.size == 0 or interspike_end.size == 0:
            return (np.nan, np.nan)
        # First interspike_start must be before first interspike_end
        if interspike_end[-1] < interspike_start[-1]:
            interspike_start = interspike_start[:-1]
        if interspike_start.size == 0:
            return (np.nan, np.nan)
        # Last interspike_start must be before last interspike_end
        if interspike_end[0] < interspike_start[0]:
            interspike_end = interspike_end[1:]

        assert interspike_start.size == interspike_end.size

        interspike_lengths = [interspike_end[i]-interspike_start[i] for i in range(interspike_start.size)]
        mean_interspike_interval = np.mean(interspike_lengths)
        interspike_variation = np.std(interspike_lengths)/mean_interspike_interval

        return (mean_interspike_interval, interspike_variation)

    def time(self, cell):
        bin_fast = self.__binarized_fast[cell]
        time = {}
        stimulation_start = self.__settings["Stimulation [frame]"][0]
        stimulation_end = self.__settings["Stimulation [frame]"][1]

        time["plateau_start"] = self.__activity[cell][0] - stimulation_start/self.__sampling
        time["plateau_end"] = self.__activity[cell][1] - stimulation_end/self.__sampling

        fast_peaks = self.__search_sequence(bin_fast[stimulation_start:], [0,1])
        if len(fast_peaks) < 3:
            time["spike_start"] = np.nan
        else:
            time["spike_start"] = (np.mean(fast_peaks[:3]))/self.__sampling

        return time


    def node_degree(self, cell):
        if self.__networks is False:
            raise ValueError("Network is not built.")
        return self.__networks.node_degree(cell)

    def clustering(self, cell):
        if self.__networks is False:
            raise ValueError("Network is not built.")
        return self.__networks.clustering(cell)

    def nearest_neighbour_degree(self, cell):
        if self.__networks is False:
            raise ValueError("Network is not built.")
        return self.__networks.nearest_neighbour_degree(cell)

# ----------------------------- ANALYSIS METHODS ----------------------------- #
# ----------------------------- Spikes vs phases ----------------------------- #

    def spikes_vs_phase(self, mode="normal"):
        phases = np.arange((np.pi/3 - np.pi/6)/2, 2*np.pi, np.pi/6)
        spikes = np.zeros((self.__cells, 12))

        # Iterate through cells
        for cell in range(self.__cells):
            start = int(self.__activity[cell][0]*self.__sampling)
            stop = int(self.__activity[cell][1]*self.__sampling)

            bin_slow = self.__binarized_slow[cell][start:stop]
            bin_fast = self.__binarized_fast[cell][start:stop]

            # Iterate through phases (1–12)
            for phase in range(1,13):
                # Bool array with True at slow phase:
                slow_isolated = bin_slow == phase

                # Bool array with True at fast spike:
                spike_indices = self.__search_sequence(bin_fast, [0,1]) + 1
                fast_unitized = np.zeros(len(bin_fast))
                fast_unitized[spike_indices] = 1

                # Bool array with True at fast spike AND slow phase
                fast_isolated_unitized = np.logical_and(slow_isolated, fast_unitized)

                # Append result
                spikes[cell, phase-1] = np.sum(fast_isolated_unitized)

        if mode == "normal":
            result = np.sum(spikes, axis=0)
        elif mode == "separate":
            result = spikes
        return (phases, result)

# ------------------------- Correlation vs distance -------------------------- #

    def correlation_vs_distance(self):
        A_dst = self.__distances_matrix()
        distances = []
        correlations_slow = []
        correlations_fast = []
        for cell1 in range(self.__cells):
            for cell2 in range(cell1):
                distances.append(A_dst[cell1, cell2])
                corr_slow = np.corrcoef(self.__filtered_slow[cell1],
                                        self.__filtered_slow[cell2])[0,1]
                corr_fast = np.corrcoef(self.__filtered_fast[cell1],
                                        self.__filtered_fast[cell2])[0,1]
                correlations_slow.append(corr_slow)
                correlations_fast.append(corr_fast)
        return (distances, correlations_slow, correlations_fast)
